module TalltyImportExport
  class Export
    attr_reader :klass, :context

    def initialize(klass)
      @klass = klass
      @context = Context.new({})
    end

    # [
    #   { key: 'entry_user_name', name: '被考核人', chain: [:entry, :user_name] },
    #   { key: 'entry_user_code', name: '被考核人工号' },
    #   { key: 'entry_user_department_name', name: '被考核人部门' },
    #   { key: 'dimension_name', name: '考核维度' },
    #   { key: 'user_name', name: '考核人' },
    #   { key: 'user_code', name: '考核人工号' },
    #   { key: 'user_department_name', name: '考核人部门' },
    #   { key: 'state', name: '考核状态', method: :state_zh },
    #   { key: 'score', name: '考核分', source: true },
    #   { key: 'model_key', name: '关键字', list: [ { key: 'model_key', name: '新属性' } ]},
    # ]
    # export_headers_result / headers
    # key: 属性的英文名，可以支持user.name这样的方式
    # name: 属性的中文名
    # attr_type: 属性的类型
    # format: excel是否需要特定的格式，目前主要是类似于身份证号，可以用string
    # method: 导出时本地调用的方法
    # chain: 导出时对象属性通过链式调用
    # index: 数组方式，需要嵌套拿到里面的
    # merge: true/false，默认false，某一列，如果上下行的内容相同，则直接合并单元格
    # json: model_payload，代表存储在某一个列中
    # select: [{ label: '已报备', value: 'submitted'}, ...]，需要转换的枚举类型
    # source: true，如果source为true，代表从association_record 进行属性查询
    # proc: proc或者lamda，支持call，传入 record 和 context
    # list: 对于list布局的，进行嵌套，生成子表格

    def export_xlsx(records, **options)
      records = with_scope records
      process_options(options)

      options = options.with_indifferent_access

      Axlsx::Package.new do |pack|
        pack.use_shared_strings = true
        workbook = pack.workbook

        if @group_by.present?
          if records.is_a?(Array)
            records.group_by { |record| record.send(@group_by) }.each do |key, group_records|
              next unless key.present?

              @group_key = key
              export_workbook workbook, group_records, **options
            end
          else
            records.group(@group_by).count.keys.each do |key|
              next unless key.present?

              @group_key = key
              export_workbook workbook, records.ransack("#{@group_where}" => key).result, **options
            end
          end
        else
          export_workbook workbook, records, **options
        end

        # 是否导出详细子表
        # export_association_headers的结构，是association结合
        if respond_to?(:export_association_headers)
          records.each do |record|
            context.record = record
            _export_association_headers = export_association_headers(**options)
            Array.wrap(_export_association_headers).each do |association_headers|
              association_headers = association_headers.with_indifferent_access
              association_records = record.send(association_headers[:association])
              sheet_name = if association_headers[:sheet_name].present?
                             association_headers[:sheet_name].respond_to?(:call) ?
                               association_headers[:sheet_name].call(record, context) :
                               association_headers[:sheet_name]
                           end
              export_workbook(
                workbook,
                association_records,
                sheet_name: sheet_name || record.try(:name),
                export_headers: association_headers[:headers],
                **options,
              )
            end
          end
        end

        file_path = File.join(Rails.root, 'public', 'export')
        # file_path = File.join('/Users/mushu/', 'export')
        FileUtils.mkdir_p(file_path) unless Dir.exist?(file_path)
        file_name = "#{Time.now.strftime('%Y%m%d%H%M%S')}#{@filename}.xlsx"
        file_path_with_name = File.join(file_path, file_name)
        pack.serialize(file_path_with_name)
        return file_path_with_name
      end
    end

    def export_workbook(workbook, association_records, sheet_name: nil, export_headers: nil, **options)
      # excel导出样式
      alignment = { vertical: :center, horizontal: :center }
      border    = { color: '969696', style: :thin }
      title1 = workbook.styles.add_style(alignment: alignment, border: border, sz: 12, b: true)
      title2 = workbook.styles.add_style(alignment: alignment, border: border, bg_color: '2a5caa', sz: 12, fg_color: 'fffffb')
      title3 = workbook.styles.add_style(alignment: alignment.merge(wrap_text: true), border: border, sz: 10)
      headers = export_headers || export_headers_result(**options)

      _sheet_name = respond_to?(:sheet_name) ? sheet_name : nil
      _sheet_name ||= sheet_name

      workbook.add_worksheet(name: _sheet_name) do |sheet|
        if respond_to?(:first_header)
          row_index = Axlsx.col_ref(headers.size - 1)
          sheet.merge_cells("A1:#{row_index}1")
          sheet.add_row [first_header], style: title1, height: 30
        end

        sheet.add_row headers.map { |header| header[:name] }, style: title2, height: 25

        last_row = nil
        merge_column_hash = {}
        first_content_row_index = respond_to?(:first_header) ? 2 : 1

        index = 0
        association_records.each do |association_record|
          records = @each_method.present? ?
            (try_method(association_record, @each_method) || [nil]) :
            [association_record]

          records.each do |record|
            row = []
            formats = []
            index += 1
            headers.each_with_index do |header, col_index|
              _data = header[:source] ?
                handle_data(association_record, header, index) :
                handle_data(record, header, index)

              if header[:merge].present? && last_row.present? && _data == last_row[col_index]
                # 这里使用二维数组，每个数组里都是列内容相同的各行
                merge_column_hash[col_index] ||= []
                if merge_column_hash[col_index].last&.last == index + first_content_row_index - 1 - 1
                  # 说明内容和上面的是延续的，继续加入之前的数组
                  merge_column_hash[col_index].last << index + first_content_row_index - 1
                else
                  merge_column_hash[col_index] << [index + first_content_row_index - 1 - 1, index + first_content_row_index - 1]
                end
              end
              row.push(_data)
              formats.push(header[:format]&.to_sym || (_data.is_a?(String) ? :string : nil))
            end
            sheet.add_row row, style: title3, height: @row_height, types: formats
            last_row = row
          end
        end
        # 需要根据column进行多行的内容合并
        if merge_column_hash.present?
          merge_column_hash.each do |col_index, row_arr|
            row_arr.each do |arr|
              sheet.merge_cells(
                Axlsx.cell_r(col_index, arr.first) + ':' + Axlsx.cell_r(col_index, arr.last),
              )
            end
          end
        end
        sheet.column_widths(*headers.map { |header| (header[:width] || @width).to_f })
      end
    end

    # 地处整体配置信息
    def export_options
      {}
    end

    def process_options(options = {})
      options = export_options.merge(options).with_indifferent_access

      @row_height ||= options.delete(:row_height) || 25
      @width ||= (options.delete(:width) || 20).to_f
      @filename ||= options.delete(:filename)
      @group_by ||= options.delete(:group_by)
      @group_where = "#{@group_by}_eq" if @group_by.present?
      @headers ||= options.delete(:headers)
      @each_method ||= options.delete(:each_method)
      @params = options
      context.params = @params
    end

    def with_scope(records)
      records
    end

    def export_headers_result(**options)
      if @headers.present? && @group_key.blank?
        headers_hash = @headers.to_h { |header| [header.with_indifferent_access[:key], header] }.with_indifferent_access
        export_headers(**options.symbolize_keys).select do |_header|
          _header.with_indifferent_access[:key].to_s.in?(headers_hash.keys)
        end.map do |_header|
          _header = _header.with_indifferent_access
          _header.merge(headers_hash[_header[:key]].delete_if { |_k, v| v.blank? })
        end
      else
        @headers = export_headers(**options.symbolize_keys)
      end
    end

    def export_headers(**args)
      @headers || klass.try(:headers) || klass.try(:model_headers)
    end

    # 处理一个记录的数据
    def handle_data(record, header, index = 0, **opts)
      data =
        if header[:key] == '_index'
          index
        elsif header[:method].present?
          send(header[:method], record, header)
        elsif header[:chain].present?
          try_chain(record, header[:chain])
        elsif header[:proc] && header[:proc].respond_to?(:call)
          header[:proc].call(record, context)
        else
          try_method(record, header[:key], prefix: [header[:prefix], header[:json]].compact.join('.'))
        end
      data = handle_format(data, header, **opts)
      data = handle_data_type(data, **opts)
      data = handle_select(data, header, **opts)
    rescue StandardError
      ''
    end

    def try_chain(record, arr)
      arr.reduce(record) do |r, m|
        if r.is_a?(Array)
          r.try(m) || r.try(:[], m.to_i)
        elsif r.is_a?(Hash)
          r.try(:[], m) || r.try(:[], m.to_sym)
        else
          data = begin
            r.try(m)
          rescue StandardError
            nil
          end
          if data.nil?
            data = begin
              r.try(:[], m)
            rescue StandardError
              nil
            end
          end
          if data.nil?
            data = begin
              r.try(:[], m.to_sym)
            rescue StandardError
              nil
            end
          end
          data
        end
      end
    end

    def try_method(record, method, prefix: nil)
      prefix_arr = prefix.to_s.split(/\./)
      arr = method.to_s.split(/\./)
      try_chain record, prefix_arr + arr
    end

    # 根据数据类型 attr_type 进行数据的格式化
    def handle_format(data, header, **opts)
      case header[:attr_type].to_s
      when 'string'
        data.to_s
      when 'datetime'
        data ? data.in_time_zone.strftime('%F %H:%M') : nil
      when 'date'
        data ? data.in_time_zone.strftime('%F') : nil
      when 'file'
        if data.is_a?(Array) && !opts[:keep_array]
          data.map do |item|
            url = item.dig('url') || item.dig('downloadPath')
            url.present? ?
              "#{url}?fileName=#{URI.encode_www_form_component(item.dig('fileName'))}&fileSize=#{URI.encode_www_form_component(item.dig('fileSize'))}&mimeType=#{URI.encode_www_form_component(item.dig('mimeType') || item.dig('fileType'))}" :
              nil
          end.compact.join("\n")
        else
          data
        end
      else
        data
      end
    end

    def handle_data_type(data, **opts)
      if data.is_a?(Time)
        data.in_time_zone.strftime('%F %H:%M')
      elsif data.is_a?(Date)
        data.in_time_zone.strftime('%F')
      elsif data.is_a?(TrueClass)
        '是'
      elsif data.is_a?(FalseClass)
        '否'
      elsif data.is_a?(Array) && !opts[:keep_array]
        if data.first.is_a?(Hash) && data.first&.dig('url').present?
          # 文件array
          data.map do |item|
            item.dig('url')
          end.compact.join("\n")
        else
          data.map(&:to_s).compact.join("\n")
        end
      else
        data
      end
    end

    def handle_select(data, origin_header, **opts)
      header = origin_header.deep_symbolize_keys
      if header[:select].present?
        select_option = header[:select].find { |option| option[:value].to_s == data.to_s }
        select_option.present? ? select_option[:label] : data
      else
        data
      end
    end
  end
end
