module TalltyImportExport
  class Import
    require 'roo'
    require 'roo-xls'
    attr_reader :klass, :context, :primary_keys, :associations

    def initialize(klass)
      @klass = klass
      @context = Context.new({})
    end

    # key: 属性的英文名
    # name: 属性的中文名
    # attr_type: 属性的类型
    # format: excel是否需要特定的格式，目前主要是类似于身份证号，可以用string
    # convert: 导入时候，把excel的内容转换成导入时候代码逻辑需要的内容
    # primary_key: 是否是主键
    # json: model_payload，代表存储在某一个列中
    # finder: 执行代码，可以查找到相关的对象
    # proc: proc或者lamda，支持call，传入 val 和 context
    # skip: 用来综合使用的数据，但是不导入

    # xlsx_file 为 file path or file object or TalltyImportExport::Excel.new
    def import_xlsx(xlsx_file, associations, **options)
      process_xlsx_line_info(xlsx_file, associations, **options) do |line_info, associations|
        process_line_info(line_info, associations)
      end
    end

    # 导出excel的空模版
    def export_template_xlsx(**options)
      origin_options = options.dup
      process_options(options)

      Axlsx::Package.new do |pack|
        pack.use_shared_strings = true
        workbook = pack.workbook

        export_workbook workbook, **origin_options

        file_path = File.join(Rails.root, 'public', 'import')
        FileUtils.mkdir_p(file_path) unless Dir.exist?(file_path)
        file_name = "#{Time.now.strftime('%Y%m%d%H%M%S')}#{@filename}.xlsx"
        file_path_with_name = File.join(file_path, file_name)
        pack.serialize(file_path_with_name)
        return file_path_with_name
      end
    end

    def export_workbook(workbook, **options)
      export_instance = TalltyImportExport::Export.new(@klass)
      opts = options.symbolize_keys.merge({ headers: @headers })
      # Rails.logger.info opts
      export_instance.process_options(opts)
      export_instance.export_workbook(workbook, [], **opts)
    end

    def exchange_to_ids(xlsx_file, associations, **options)
      errors = []
      ids = []
      process_xlsx_line_info(xlsx_file, associations, **options) do |line_info, associations|
        ids << exchange_line_info_to_id(line_info, associations, errors)
      end

      if errors.any?
        error_msg = errors.map do |line_info|
          "【#{@primary_keys.map { |key| line_info[key] }.join(' - ')}】"
        end.join(' ')
        raise RecordNotFountError, "以下内容未找到: #{error_msg}"
      end
      ids.compact.uniq
    end

    def process_xlsx_line_info(xlsx_file, associations, **options)
      @associations = associations
      # 先处理获取出来Excel每行的数据, line_info
      process_options(options)

      if xlsx_file.is_a?(TalltyImportExport::Excel)
        xlsx_file.rows.each_with_excel_hash(@excel_hash) do |line_info|
          yield line_info.with_indifferent_access, associations
        end
      else
        file_path = xlsx_file.is_a?(String) ? xlsx_file : xlsx_file.path
        xlsx = ::Roo::Spreadsheet.open(file_path, extension: File.extname(file_path) == '.xls' ? :xls : :xlsx)
        xlsx.each_with_pagename do |_sheetname, sheet|
          sheet.each(**@excel_hash).with_index do |line_info, index|
            next if index == 0

            yield line_info.with_indifferent_access, associations
          end
        end
      end
    end

    def import_data(data, associations, **options)
      process_options(options)
      TalltyImportExport::Excel::Rows.new(data).each_with_excel_hash(@excel_hash) do |line_info|
        process_line_info(line_info.with_indifferent_access, associations)
      end
    end

    def tallty_excel
      TalltyImportExport::Excel
    end

    def process_options(options)
      options = import_options.merge(options).with_indifferent_access
      # Rails.logger.info options
      @headers = options.delete(:headers) || import_headers(**options.symbolize_keys)
      @primary_keys = options.delete(:primary_keys) || @headers.map { |header| header[:primary_key] ? header[:key].to_sym : nil }.compact
      @skip_keys = options.delete(:skip_keys) || @headers.map { |header| header[:skip] ? header[:key].to_sym : nil }.compact
      @params = options
      context.params = @params

      @excel_hash = @headers.each_with_object({}) do |header, h|
        h[header[:key].to_sym] = header[:name]
      end

      options
    end

    def process_line_info(line_info, associations)
      # 去除空行内容
      return unless line_info.values.any?(&:present?)

      context.line_info = line_info
      # 转换处理导入的数据格式
      line_info = convert_data(line_info)

      # 处理每行对于导入的动作,处理line_info
      return unless valid?(line_info)

      import_record(line_info, associations)
      context.last_line_info = line_info
    end

    def convert_data(line_info)
      info = line_info.with_indifferent_access
      import_headers_result.each_with_object({}.with_indifferent_access) do |header, h|
        k = header[:key]
        k_arr = k.to_s.split(/\./)
        v = info[k]
        # header[:convert] = handle_xxx
        # handle_xxx(val, processing_line_info, raw_line_info)
        val = header[:convert] ? send(header[:convert], v, h, info) : v
        val.strip! if val.is_a?(String)

        val = handle_attr_type(val, header, line_info)

        if header[:json]
          h[header[:json]] ||= {}
          h[header[:json]][k] = val
        elsif header[:proc] && header[:proc].respond_to?(:call)
          h[k.to_sym] = header[:proc].call(val, context)
        elsif header[:finder]
          # $SAFE = 2
          h[k.to_sym] = eval header[:finder]
        elsif k_arr.size > 1
          # hash设置
          attr = k_arr.first
          val = { k_arr[1..-1].join('.') => val }.to_nested_hash
          h[attr.to_sym] ||= {}
          h[attr.to_sym].deep_merge! val
        else
          h[k.to_sym] = val
        end
      end.with_indifferent_access
    end

    def handle_attr_type val, header, line_info
      case header[:attr_type].to_s
      when 'file'
        if val.is_a?(String)
          val.split("\n").map do |url|
            params = url.include?('?') ? CGI::parse(url.split('?').last) : {}
            {
              url: url,
              fileName: params['fileName']&.first,
              mimeType: params['mimeType']&.first,
              fileSize: params['fileSize']&.first&.to_i,
            }
          end
        else
          val
        end
      else
        val
      end
    end

    # 通过转换后，数据是否合法，如果不合法，则直接跳过不处理这个数据
    def valid?(line_info)
      true
    end

    # TODO: 这里，对于import_headers，后面还是不要传参数了
    # 需要合并proc，前端没有办法把proc传过来
    def import_headers_result
      if @headers.present?
        headers_hash = @headers.to_h { |header| [header.with_indifferent_access[:key], header] }.with_indifferent_access
        import_headers(**context.params.symbolize_keys).select do |_header|
          _header.with_indifferent_access[:key].to_s.in?(headers_hash.keys)
        end.map do |_header|
          _header = _header.with_indifferent_access
          _header.merge(headers_hash[_header[:key]].delete_if { |_k, v| v.blank? })
        end
      else
        @headers = import_headers
      end
    rescue StandardError
      @headers
    end

    def import_options
      {}
    end

    def import_headers(**args)
      @headers || klass.try(:headers) || klass.try(:model_headers) || (raise ArgumentError, 'missing import_headers')
    end

    # # 只保留 key, name, json, 合并到 import_header
    # def headers= val
    #   if val.empty?
    #     @headers = import_headers_result.map { |header| header.with_indifferent_access }
    #     return
    #   end
    #
    #   key_to_coming_header = val.reduce({}) do |out, header|
    #     out[header.with_indifferent_access[:key].to_sym] = header.with_indifferent_access
    #     out
    #   end
    #
    #   result = []
    #   val.map do |header|
    #     if (exist_header = import_headers_result.find { |model_header| model_header[:key] === header[:key] })
    #       result.push(exist_header.merge(header.compact))
    #     else
    #       result.push(header)
    #     end
    #   end
    #
    #   @headers = result
    # end

    def skip(val, processing_line_info, raw_line_info)
      # do nothing there, use for header[:convert]
    end

    ###  这个方法是可以由复杂业务进行重载的 ###
    def import_record(line_info, associations)
      if primary_keys.present?
        _record = associations.find_or_initialize_by(line_info.clone.extract!(*primary_keys))
        _record.update!(line_info.clone.except!(*primary_keys, *@skip_keys))
        _record
      else
        associations.create!(line_info.clone.except!(*@skip_keys))
      end
    end

    def exchange_line_info_to_id(line_info, associations, errors)
      # 去除空行内容
      return unless line_info.values.any?(&:present?)

      context.line_info = line_info

      return unless primary_keys.present?

      ids = associations.where(line_info.clone.extract!(*primary_keys)).pluck(:id)

      context.last_line_info = line_info

      errors << line_info unless ids[0]
      ids[0]
    end

    class RecordNotFountError < StandardError
      attr_accessor :message

      def initialize(message)
        @message = message
        super()
      end
    end
  end
end
